# Copyright 2012 Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test the client."""
import mock

from uuid import UUID

from twisted.internet import defer

from silverberg.client import CQLClient, ConsistencyLevel, TestingCQLClient

from silverberg.cassandra import ttypes, Cassandra

from silverberg.test.util import BaseTestCase


class MockClientTests(BaseTestCase):

    """Test the client."""

    def setUp(self):
        """Setup the mock objects for the tests."""
        self.endpoint = mock.Mock()
        self.client_proto = mock.Mock(Cassandra.Client)
        self.twisted_transport = mock.Mock()

        self.mock_results = ttypes.CqlResult(type=ttypes.CqlResultType.INT, num=1)

        self.client_proto.set_keyspace.return_value = defer.succeed(None)
        self.client_proto.login.return_value = defer.succeed(None)
        self.client_proto.describe_version.return_value = defer.succeed('1.2.3')

        def _execute_cql3_query(*args, **kwargs):
            return defer.succeed(self.mock_results)

        self.client_proto.execute_cql3_query.side_effect = _execute_cql3_query

        def _connect(factory):
            wrapper = mock.Mock()
            wrapper.transport = self.twisted_transport
            wrapper.wrapped.client = self.client_proto
            return defer.succeed(wrapper)

        self.endpoint.connect.side_effect = _connect

    def test_disconnect_on_cancel(self):
        """
        If allowed, cancellation of running query will also try to disconnect
        the TCP connection
        """
        self.client_proto.execute_cql3_query.side_effect = lambda *_: defer.Deferred()
        client = CQLClient(self.endpoint, 'abc', disconnect_on_cancel=True)
        client.disconnect = mock.Mock()

        d = client.execute('query', {}, ConsistencyLevel.ONE)

        self.assertNoResult(d)
        self.assertFalse(client.disconnect.called)

        d.cancel()
        self.failureResultOf(d, defer.CancelledError)
        client.disconnect.assert_called_one_with()

    def test_disconnect_on_cancel_returns_correct_value(self):
        """
        with disconnect_on_cancel=True, the value from execute_cql3_query is
        returned before cancellation
        """
        exec_d = defer.Deferred()
        self.client_proto.execute_cql3_query.side_effect = lambda *_: exec_d
        client = CQLClient(self.endpoint, 'abc', disconnect_on_cancel=True)
        client.disconnect = mock.Mock()

        d = client.execute('query', {}, ConsistencyLevel.ONE)

        self.assertNoResult(d)
        self.assertFalse(client.disconnect.called)
        exec_d.callback(self.mock_results)
        self.assertEqual(self.successResultOf(d), 1)
        self.assertFalse(client.disconnect.called)

    def test_no_disconnect_on_cancel(self):
        """
        If not given, cancellation of running query should not try to disconnect
        the TCP connection
        """
        self.client_proto.execute_cql3_query.side_effect = lambda *_: defer.Deferred()
        client = CQLClient(self.endpoint, 'abc', disconnect_on_cancel=False)
        client.disconnect = mock.Mock()

        d = client.execute('query', {}, ConsistencyLevel.ONE)

        self.assertNoResult(d)
        self.assertFalse(client.disconnect.called)

        d.cancel()
        self.failureResultOf(d, defer.CancelledError)
        self.assertFalse(client.disconnect.called)

    def test_disconnect(self):
        """
        When disconnect is called, the on demand thrift client is disconnected
        """
        client = CQLClient(self.endpoint, 'blah')
        self.assertFired(client.describe_version())
        client.disconnect()
        self.twisted_transport.loseConnection.assert_called_once_with()

    def test_login(self):
        """Test that login works as expected."""
        client = CQLClient(self.endpoint, 'blah', 'groucho', 'swordfish')

        d = client.describe_version()
        self.assertEqual(self.assertFired(d), '1.2.3')
        self.client_proto.describe_version.assert_called_once_with()

        self.client_proto.set_keyspace.assert_called_once_with('blah')

        creds = {'username': 'groucho', 'password': 'swordfish'}
        authreq = ttypes.AuthenticationRequest(creds)
        self.client_proto.login.assert_called_once_with(authreq)

    def test_bad_keyspace(self):
        """Ensure that a bad keyspace results in an errback."""
        self.client_proto.set_keyspace.return_value = defer.fail(ttypes.NotFoundException())
        client = CQLClient(self.endpoint, 'blah')

        d = client.describe_version()
        self.assertFailed(d, ttypes.NotFoundException)
        self.client_proto.set_keyspace.assert_called_once_with('blah')

    def test_describe_version(self):
        """Connect and check the version."""
        client = CQLClient(self.endpoint, 'blah')

        d = client.describe_version()
        self.assertEqual(self.assertFired(d), '1.2.3')
        self.assertEqual(self.client_proto.describe_version.call_count, 1)
        self.client_proto.set_keyspace.assert_called_once_with('blah')

    def test_unsupported_types_are_returned_as_bytes(self):
        """
        When a table includes a column of a type that is not explicitly
        supported we should return the raw bytes instead of attempting to
        unmarshal the data.
        """
        mock_rows = [ttypes.CqlRow(
            key='',
            columns=[
                ttypes.Column(
                    name='an_unknown_type',
                    value="\x00\x01")])]

        self.mock_results = ttypes.CqlResult(
            type=ttypes.CqlResultType.ROWS,
            rows=mock_rows,
            schema=ttypes.CqlMetadata(value_types={'an_unknown_type': 'an.unknown.type'}))

        client = CQLClient(self.endpoint, 'blah')
        d = client.execute("SELECT * FROM blah", {}, ConsistencyLevel.ONE)
        results = self.assertFired(d)

        self.assertEqual(results, [{'an_unknown_type': '\x00\x01'}])

    def test_cql_value(self):
        """
        Test that a CQL response that is an integer value is
        processed correctly (e.g. SELECT COUNT).

        """
        self.mock_results = ttypes.CqlResult(type=ttypes.CqlResultType.INT, num=1)
        client = CQLClient(self.endpoint, 'blah')

        d = client.execute("SELECT :sel FROM test_blah", {"sel": "blah"}, ConsistencyLevel.ONE)
        self.assertEqual(self.assertFired(d), 1)
        self.client_proto.execute_cql3_query.assert_called_once_with("SELECT 'blah' FROM test_blah", 2,
                                                                     ConsistencyLevel.ONE)
        self.client_proto.set_keyspace.assert_called_once_with('blah')

    def test_cql_array(self):
        """Test that a full CQL response (e.g. SELECT) works."""
        expected = [{"foo": "{P}"}]

        mockrow = [ttypes.CqlRow(key='blah', columns=[ttypes.Column(name='foo', value='{P}')])]
        self.mock_results = ttypes.CqlResult(
            type=ttypes.CqlResultType.ROWS,
            rows=mockrow,
            schema=ttypes.CqlMetadata(value_types={'foo': 'org.apache.cassandra.db.marshal.UTF8Type'}))
        client = CQLClient(self.endpoint, 'blah')

        d = client.execute("SELECT :sel FROM test_blah", {"sel": "blah"}, ConsistencyLevel.ONE)
        self.assertEqual(self.assertFired(d), expected)
        self.client_proto.execute_cql3_query.assert_called_once_with("SELECT 'blah' FROM test_blah", 2,
                                                                     ConsistencyLevel.ONE)
        self.client_proto.set_keyspace.assert_called_once_with('blah')

    def test_cql_array_deserial(self):
        """Make sure that values that need to be deserialized correctly are."""
        expected = [{"fff": 1222}]

        mockrow = [ttypes.CqlRow(key='blah', columns=[ttypes.Column(name='fff', value='\x04\xc6')])]
        self.mock_results = ttypes.CqlResult(type=ttypes.CqlResultType.ROWS,
                                             rows=mockrow,
                                             schema=ttypes.CqlMetadata(value_types={
                                                 'fff': 'org.apache.cassandra.db.marshal.IntegerType'
                                             }))
        client = CQLClient(self.endpoint, 'blah')

        d = client.execute("SELECT * FROM :tablename;", {"tablename": "blah"}, ConsistencyLevel.ONE)
        self.assertEqual(self.assertFired(d), expected)
        self.client_proto.execute_cql3_query.assert_called_once_with("SELECT * FROM 'blah';", 2,
                                                                     ConsistencyLevel.ONE)
        self.client_proto.set_keyspace.assert_called_once_with('blah')

    def test_cql_list_deserial(self):
        expected = [{'fff': ['ggg', 'hhh']}]

        mockrow = [ttypes.CqlRow(key='blah',
                                 columns=[ttypes.Column(name='fff',
                                                        value='\x00\x02\x00\x03ggg\x00\x03hhh')])]

        list_type = 'org.apache.cassandra.db.marshal.ListType'
        text_type = 'org.apache.cassandra.db.marshal.UTF8Type'
        text_list_type = list_type + '(' + text_type + ')'

        self.mock_results = ttypes.CqlResult(
            type=ttypes.CqlResultType.ROWS,
            rows=mockrow,
            schema=ttypes.CqlMetadata(value_types={'fff': text_list_type}))
        client = CQLClient(self.endpoint, 'blah')

        d = client.execute("SELECT * FROM :tablename;", {"tablename": "blah"}, ConsistencyLevel.ONE)
        self.assertEqual(self.assertFired(d), expected)
        self.client_proto.execute_cql3_query.assert_called_once_with("SELECT * FROM 'blah';", 2,
                                                                     ConsistencyLevel.ONE)
        self.client_proto.set_keyspace.assert_called_once_with('blah')

    def test_cql_None_not_deserialized(self):
        """
        If the value is None, it is not deserialized at all.
        """
        raw_rows = [ttypes.CqlRow(
            key='blah', columns=[ttypes.Column(name='fff', value=None)])]
        schema = ttypes.CqlMetadata(value_types={
            'fff': 'org.apache.cassandra.db.marshal.AlwaysFailType'})

        client = CQLClient(self.endpoint, 'blah')

        always_blow_up = mock.Mock(spec=[], side_effect=Exception)

        rows = client._unmarshal_result(schema, raw_rows, {
            'org.apache.cassandra.db.marshal.AlwaysFailType': always_blow_up
        })

        self.assertEqual(rows, [{'fff': None}])
        self.assertEqual(always_blow_up.call_count, 0)

    def test_cql_insert(self):
        """Test a mock CQL insert with a VOID response works."""
        expected = None

        self.mock_results = ttypes.CqlResult(type=ttypes.CqlResultType.VOID)
        client = CQLClient(self.endpoint, 'blah')

        d = client.execute("UPDATE blah SET 'key'='frr', 'fff'=1222 WHERE KEY='frr'", {},
                           ConsistencyLevel.ONE)
        self.assertEqual(self.assertFired(d), expected)
        self.client_proto.execute_cql3_query.assert_called_once_with(
            "UPDATE blah SET 'key'='frr', 'fff'=1222 WHERE KEY='frr'",
            2, ConsistencyLevel.ONE)
        self.client_proto.set_keyspace.assert_called_once_with('blah')

    def test_cql_insert_vars(self):
        """Test that a CQL insert that has variables works."""
        expected = None

        self.mock_results = ttypes.CqlResult(type=ttypes.CqlResultType.VOID)
        client = CQLClient(self.endpoint, 'blah')

        d = client.execute("UPDATE blah SET 'key'='frr', 'fff'=:val WHERE KEY='frr'", {"val": 1234},
                           ConsistencyLevel.ONE)
        self.assertEqual(self.assertFired(d), expected)
        self.client_proto.execute_cql3_query.assert_called_once_with(
            "UPDATE blah SET 'key'='frr', 'fff'=1234 WHERE KEY='frr'",
            2, ConsistencyLevel.ONE)
        self.client_proto.set_keyspace.assert_called_once_with('blah')

    def test_cql_sequence(self):
        """
        Test a sequence of operations results in only one handshake
        but two requests.

        """
        expected = [{"foo": "{P}"}]

        mockrow = [ttypes.CqlRow(key='blah', columns=[ttypes.Column(name='foo', value='{P}')])]
        self.mock_results = ttypes.CqlResult(
            type=ttypes.CqlResultType.ROWS, rows=mockrow,
            schema=ttypes.CqlMetadata(
                value_types={'foo': 'org.apache.cassandra.db.marshal.UTF8Type'}))
        client = CQLClient(self.endpoint, 'blah')

        def _cqlProc(r):
            return client.execute("SELECT :sel FROM test_blah", {"sel": "blah"},
                                  ConsistencyLevel.ONE)

        d = client.execute("SELECT :sel FROM test_blah", {"sel": "ffh"},
                           ConsistencyLevel.ONE)
        d.addCallback(_cqlProc)
        self.assertEqual(self.assertFired(d), expected)
        self.client_proto.execute_cql3_query.assert_any_call("SELECT 'blah' FROM test_blah", 2,
                                                             ConsistencyLevel.ONE)
        self.client_proto.execute_cql3_query.assert_any_call("SELECT 'ffh' FROM test_blah", 2,
                                                             ConsistencyLevel.ONE)
        self.client_proto.set_keyspace.assert_called_once_with('blah')

    def test_cql_result_metadata(self):
        """
        execute should use the metadata included with the CqlResult for
        deserializing values.
        """
        expected = [{"foo": UUID('114b8328-d1f1-11e2-8683-000c29bc9473')}]

        mockrow = [
            ttypes.CqlRow(
                key='blah',
                columns=[
                    ttypes.Column(
                        name='foo',
                        value='\x11K\x83(\xd1\xf1\x11\xe2\x86\x83\x00\x0c)\xbc\x94s')])]

        self.mock_results = ttypes.CqlResult(
            type=ttypes.CqlResultType.ROWS,
            rows=mockrow,
            schema=ttypes.CqlMetadata(value_types={
                'foo': 'org.apache.cassandra.db.marshal.TimeUUIDType'}))

        client = CQLClient(self.endpoint, 'blah')
        d = client.execute("SELECT * FROM blah;", {}, ConsistencyLevel.ONE)
        self.assertEqual(self.assertFired(d), expected)


class MockTestingClientTests(MockClientTests):
    """
    Test the conveniences provided by the testing client
    """

    def test_transport_exposed(self):
        """
        The transport exposed is the underlying twisted transport, if it exists
        """
        client = TestingCQLClient(self.endpoint, 'meh')
        self.assertEqual(client.transport, None)  # has not connected yet
        self.assertFired(client.describe_version())
        self.assertIs(client.transport, self.twisted_transport)

    def test_pause(self):
        """
        When pausing, stop reading and stop writing on the transport are called
        if the transport exists.
        """
        client = TestingCQLClient(self.endpoint, 'meh')
        client.pause()
        self.assertEqual(len(self.twisted_transport.stopReading.mock_calls), 0)
        self.assertEqual(len(self.twisted_transport.stopWriting.mock_calls), 0)

        self.assertFired(client.describe_version())
        client.pause()
        self.twisted_transport.stopReading.assert_called_one_with()
        self.twisted_transport.stopWriting.assert_called_one_with()

    def test_resume(self):
        """
        When resuming, start reading and start writing on the transport are
        called if the transport exists.
        """
        client = TestingCQLClient(self.endpoint, 'meh')
        client.pause()
        self.assertEqual(len(self.twisted_transport.startReading.mock_calls),
                         0)
        self.assertEqual(len(self.twisted_transport.startWriting.mock_calls),
                         0)

        self.assertFired(client.describe_version())
        client.pause()
        self.twisted_transport.startReading.assert_called_one_with()
        self.twisted_transport.startWriting.assert_called_one_with()

# class FaultTestCase(BaseTestCase):
#     def setUp(self):
#         self.client = CqlClient(TCP4ClientEndpoint(reactor, '127.0.0.1', 9160), 'blah')

#     def test_vers(self):
#         d = self.client.describe_version()
#         def printR(r):
#             print r
#         d.addCallback(printR)
#         return d

#     def test_cql(self):
#         d = self.client.execute("SELECT * FROM blah;", {})
#         def printQ(r):
#             print r
#         d.addCallback(printQ)
#         return d



# Copyright 2012 Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

Client API library for the Silverberg Twisted Cassandra CQL interface.

"""
import re

from silverberg.cassandra import Cassandra
from silverberg.cassandra import ttypes

from twisted.internet.defer import succeed, Deferred

from silverberg.marshal import prepare, unmarshallers

from silverberg.thrift_client import OnDemandThriftClient
from silverberg.cassandra.ttypes import ConsistencyLevel


class CQLClient(object):
    """
    Cassandra CQL Client object.

    Instantiate it and it will on-demand create a connection to the Cassandra
    cluster.

    :param cass_endpoint: A twisted Endpoint
    :type cass_endpoint: twisted.internet.interfaces.IStreamClientEndpoint

    :param keyspace: A keyspace to connect to
    :type keyspace: str.

    :param user: Username to connect with.
    :type user: str.

    :param password: Password to use in conjunction with Username.
    :type password: str.

    :param bool disconnect_on_cancel: Should the underlying TCP connection be disconnected
    when execute deferred is cancelled?

    Upon connecting, the client will authenticate (if paramaters are provided)
    and obtain the keyspace definition so that it can de-serialize properly.

    n.b. Cassandra presently doesn't have any real support for password
    authentication in the mainline as the simple access control options
    are disabled; you probably need to secure your Cassandra server using
    different methods and the password code isn't heavily tested.
    """

    def __init__(self, cass_endpoint, keyspace, user=None, password=None,
                 disconnect_on_cancel=False):
        self._client = OnDemandThriftClient(cass_endpoint, Cassandra.Client)

        self._keyspace = keyspace
        self._user = user
        self._password = password
        self._disconnect_on_cancel = disconnect_on_cancel

    def _set_keyspace(self, client):
        d = client.set_keyspace(self._keyspace)
        return d.addCallback(lambda _: client)

    def _login(self, client):
        creds = {'username': self._user, 'password': self._password}
        authreq = ttypes.AuthenticationRequest(creds)
        d = client.login(authreq)
        d.addCallback(lambda _: client)
        return d

    def _connection(self):
        def _handshake(client):
            d = succeed(client)
            if self._user and self._password:
                d.addCallback(self._login)
            d.addCallback(self._set_keyspace)
            return d

        ds = self._client.connection(_handshake)
        return ds

    def disconnect(self):
        """
        Disconnect from the cassandra cluster.  Cassandara and Silverberg do
        not require the connection to be closed before exiting.  However, this
        method may be useful if resources are constrained, or for testing
        purposes if using or injecting :class:`TestCQLClient: is impossible.

        :return: a :class:`Deferred` that fires with None when disconnected.
        """
        return self._client.disconnect()

    def describe_version(self):
        """
        Query the Cassandra server for the version.

        :returns: string -- the version tag
        """
        def _vers(client):
            return client.describe_version()

        d = self._connection()
        d.addCallback(_vers)
        return d

    def _unmarshal_result(self, schema, raw_rows, _unmarshallers):
        rows = []

        def _unmarshal_val(vtype, val):
            if val is None:
                return

            # Differentiate between regular and collection types
            # Collection types look like 'ListType(SomeCassandraType)', or 'MapType(KeyType, ValType)'
            # so try split into 1~3 parts and check if we can marshal them

            types = re.split('\(|,|\)', str(vtype).rstrip(')'))

            # Regular type
            if len(types) == 1:
                if vtype in _unmarshallers:
                    return _unmarshallers[vtype](val)

            # List/Set
            elif len(types) == 2:
                if types[0] in _unmarshallers and types[1] in _unmarshallers:
                    return _unmarshallers[types[0]](types[1], val)

            # Map
            elif len(types) == 3:
                if types[0] in _unmarshallers and types[1] in _unmarshallers \
                        and types[2] in _unmarshallers:
                    return _unmarshallers[types[0]](types[1], types[2], val)

            # XXX: We do not currently implement the full range of types.
            # So we can not unmarshal all types in which case we should just
            # return the raw bytes.
            return val

        for raw_row in raw_rows:
            row = {}
            for raw_col in raw_row.columns:
                specific = schema.value_types[raw_col.name]
                row[raw_col.name] = _unmarshal_val(specific, raw_col.value)
            rows.append(row)
        return rows

    def execute(self, query, args, consistency):
        """
        Execute a CQL query against the server.

        :param query: The CQL query to execute
        :type query: str.

        :param args: The arguments to substitute
        :type args: dict.

        :param consistency: The consistency level
        :type consistency: ConsistencyLevel

        In order to avoid unpleasant issues of CQL injection
        (Hey, just because there's no SQL doesn't mean that Little
        Bobby Tables won't mess things up for you like in XKCD #327)
        you probably want to use argument substitution instead of
        concatting strings together to build a query.

        Thus, like the official CQL driver for non-Twisted python
        that comes with the Cassandra distro, we do variable substitution.

        Example::

            d = client.execute("UPDATE :table SET 'fff' = :val WHERE "
            "KEY = :key",{"val":1234, "key": "fff", "table": "blah"})

        :returns: A Deferred that fires with either None, an int, or an
                  iterable of `{'column': value, ...}` dictionaries, depending
                  on the CQL query.  e.g. a UPDATE would return None,
                  whereas a SELECT would return an int or some rows

        Example output::

            [{"fff": 1222}]
        """
        prep_query = prepare(query, args)

        def _execute(client):
            exec_d = client.execute_cql3_query(prep_query,
                                               ttypes.Compression.NONE, consistency)
            if self._disconnect_on_cancel:
                cancellable_d = Deferred(lambda d: self.disconnect())
                exec_d.chainDeferred(cancellable_d)
                return cancellable_d
            else:
                return exec_d

        def _proc_results(result):
            if result.type == ttypes.CqlResultType.ROWS:
                return self._unmarshal_result(result.schema, result.rows,
                                              unmarshallers)
            elif result.type == ttypes.CqlResultType.INT:
                return result.num
            else:
                return None

        d = self._connection()
        d.addCallback(_execute)
        d.addCallback(_proc_results)
        return d


class TestingCQLClient(CQLClient):
    """
    Cassandra CQL Client object to be used for testing purposes.  This client
    exposes the underlying Twisted transport and provides convenience functions
    so that it can be used in trial tests.

    Instantiate it and it will on-demand create a connection to the Cassandra
    cluster.

    :param cass_endpoint: A twisted Endpoint
    :type cass_endpoint: twisted.internet.interfaces.IStreamClientEndpoint

    :param keyspace: A keyspace to connect to
    :type keyspace: str.

    :param user: Username to connect with.
    :type user: str.

    :param password: Username to connect with.
    :type password: str.

    Upon connecting, the client will authenticate (if paramaters are provided)
    and obtain the keyspace definition so that it can de-serialize properly.

    n.b. Cassandra presently doesn't have any real support for password
    authentication in the mainline as the simple access control options
    are disabled; you probably need to secure your Cassandra server using
    different methods and the password code isn't heavily tested.
    """
    @property
    def transport(self):
        """
        Get the underlying Twisted transport.
        """
        return self._client._transport

    def pause(self):
        """
        Pause the client by removing the connection from the reactor.  This is
        useful in tests if, for instance, latency is a problem and you do not
        want to disconnect and reconnect between every test.  If you do not
        disconnect and reconnect, and you do not pause and resume, then if you
        use Twisted's testing framework (``trial``), tests will fail with a
        dirty reactor warning.
        """
        if self.transport:
            self.transport.stopReading()
            self.transport.stopWriting()

    def resume(self):
        """
        Resume the client by making sure the reactor is aware of the
        connection. This is useful in tests if, for instance, latency is a
        problem and you do not want to disconnect and reconnect between every
        test.  If you do not disconnect and reconnect, and you do not pause
        and resume, then if you use Twisted's testing framework (``trial``),
        tests will fail with a dirty reactor warning.
        """
        if self.transport:
            self.transport.startReading()
            self.transport.startWriting()


__all__ = ["CQLClient", "ConsistencyLevel", "TestingCQLClient"]



# Copyright 2012 Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from twisted.python.failure import Failure


class LoggingCQLClient(object):
    """
    A logging CQL client. Every query will be timed and logged

    :param client: A `CQLClient` or `RoundRobinCassandraCluster` instance
    :param log: A bound logger that has .msg() method
    """

    def __init__(self, client, log, clock=None):
        self._client = client
        self._log = log
        if clock:
            self._clock = clock
        else:
            from twisted.internet import reactor
            self._clock = reactor

    def execute(self, query, args, consistency):
        """
        See :py:func:`silverberg.client.CQLClient.execute`
        """
        start_seconds = self._clock.seconds()

        def record_time(result):
            seconds_taken = self._clock.seconds() - start_seconds
            kwargs = dict(query=query, data=args, consistency=consistency,
                          seconds_taken=seconds_taken)
            if isinstance(result, Failure):
                self._log.msg('CQL query execution failed',
                              reason=result, **kwargs)
            else:
                self._log.msg('CQL query executed successfully', **kwargs)
            return result

        return self._client.execute(query, args, consistency).addBoth(record_time)

    def disconnect(self):
        """
        See :py:func:`silverberg.client.CQLClient.disconnect`
        """
        return self._client.disconnect()




# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lifted from the Cassandra CQL driver

import re
import struct
from uuid import UUID
from datetime import datetime
import calendar

try:
    from blist import sortedset
    assert sortedset  # a trick to get rid of E123 from pyflakes
except ImportError:
    sortedset = set

import cql

__all__ = ['prepare', 'marshal', 'unmarshal_noop', 'unmarshallers']

_param_re = re.compile(r"(?<!strategy_options)(:[a-zA-Z_][a-zA-Z0-9_]*)", re.M)

BYTES_TYPE = "org.apache.cassandra.db.marshal.BytesType"
BOOLEAN_TYPE = "org.apache.cassandra.db.marshal.BooleanType"
ASCII_TYPE = "org.apache.cassandra.db.marshal.AsciiType"
UTF8_TYPE = "org.apache.cassandra.db.marshal.UTF8Type"
INTEGER_TYPE = "org.apache.cassandra.db.marshal.IntegerType"
INTEGER32_TYPE = "org.apache.cassandra.db.marshal.Int32Type"
LONG_TYPE = "org.apache.cassandra.db.marshal.LongType"
UUID_TYPE = "org.apache.cassandra.db.marshal.UUIDType"
LEXICAL_UUID_TYPE = "org.apache.cassandra.db.marshal.LexicalType"
TIME_UUID_TYPE = "org.apache.cassandra.db.marshal.TimeUUIDType"
TIMESTAMP_TYPE = "org.apache.cassandra.db.marshal.DateType"
NEW_TIMESTAMP_TYPE = "org.apache.cassandra.db.marshal.TimestampType"
COUNTER_TYPE = "org.apache.cassandra.db.marshal.CounterColumnType"
DOUBLE_TYPE = "org.apache.cassandra.db.marshal.DoubleType"

LIST_TYPE = "org.apache.cassandra.db.marshal.ListType"
SET_TYPE = "org.apache.cassandra.db.marshal.SetType"
MAP_TYPE = "org.apache.cassandra.db.marshal.MapType"


def prepare(query, params):
    """
    For every match of the form ":param_name", call marshal
    on kwargs['param_name'] and replace that section of the query
    with the result
    """
    def repl(match):
        name = match.group(1)[1:]
        if name in params:
            return marshal(params[name])
        return ":%s" % name

    new, count = re.subn(_param_re, repl, query)

    if len(params) > count:
        raise cql.ProgrammingError("More keywords were provided "
                                   "than parameters")
    return new


def marshal(term):
    if isinstance(term, list):
        return '[' + ','.join(map(marshal, term)) + ']'
    if isinstance(term, unicode):
        return "'%s'" % __escape_quotes(term.encode('utf8'))
    elif isinstance(term, str):
        return "'%s'" % __escape_quotes(term)
    elif isinstance(term, datetime):
        # If the datetime is naive, then it is considered UTC time and stored. If it is
        # timezone-aware, then its corresponding UTC time is stored
        return str(int(calendar.timegm(term.utctimetuple()) * 1000 + term.microsecond / 1e3))
    elif isinstance(term, set) or isinstance(term, sortedset):
        # python str() produces: set([1, 2, 3]), while cassandra wants: {1, 2, 3}
        # so we have to marshal each element one by one.
        return '{%s}' % (', '.join(map(marshal, term)))
    elif term is None:
        return "null"
    else:
        return str(term)


def unmarshal_noop(bytestr):
    return bytestr


def unmarshal_bool(bytestr):
    return bytestr == "\x01"


def unmarshal_utf8(bytestr):
    return bytestr.decode("utf8")


def unmarshal_int(bytestr):
    return decode_bigint(bytestr)


def unmarshal_initializable_int(bytestr):
    """
    This is useful for counters, which may not be initialized (bytestring could be None).
    """
    if bytestr is None:
        return None
    return decode_bigint(bytestr)


_long_packer = struct.Struct('>q')

_double_packer = struct.Struct('>d')


def unmarshal_long(bytestr):
    return _long_packer.unpack(bytestr)[0]


def unmarshal_double(bytestr):
    return _double_packer.unpack(bytestr)[0]


def unmarshal_timestamp(bytestr):
    epoch = unmarshal_long(bytestr)
    return datetime.utcfromtimestamp(epoch / 1000.0)


def unmarshal_uuid(bytestr):
    return UUID(bytes=bytestr)


def unmarshal_list(objtype, bytesstr):
    result = []
    # First two bytes are an integer of list size
    numelements = unmarshal_int(bytesstr[:2])
    p = 2
    for n in range(numelements):
        # Two bytes of list element size, as an integer
        length = unmarshal_int(bytesstr[p:p + 2])
        p += 2
        # Next 'length' bytes need unmarshalling into the specific type
        value = unmarshallers[objtype](bytesstr[p:p + length])
        p += length
        result.append(value)

    return result


def unmarshal_set(objtype, bytesstr):
    result = sortedset()
    # First two bytes are an integer of list size
    numelements = unmarshal_int(bytesstr[:2])
    p = 2
    for n in range(numelements):
        # Two bytes of list element size, as an integer
        length = unmarshal_int(bytesstr[p:p + 2])
        p += 2
        # Next 'length' bytes need unmarshalling into the specific type
        value = unmarshallers[objtype](bytesstr[p:p + length])
        p += length
        result.add(value)

    return result


def unmarshal_map(keytype, valtype, bytesstr):
    result = {}
    # First two bytes are an integer of list size
    numelements = unmarshal_int(bytesstr[:2])
    p = 2
    for n in range(numelements):
        # key len
        length = unmarshal_int(bytesstr[p:p + 2])
        p += 2
        # key
        key = unmarshallers[keytype](bytesstr[p:p + length])
        p += length

        # val len
        length = unmarshal_int(bytesstr[p:p + 2])
        p += 2
        # val
        val = unmarshallers[valtype](bytesstr[p:p + length])
        p += length

        result[key] = val

    return result

unmarshallers = {BYTES_TYPE:        unmarshal_noop,
                 BOOLEAN_TYPE:      unmarshal_bool,
                 ASCII_TYPE:        unmarshal_noop,
                 UTF8_TYPE:         unmarshal_utf8,
                 INTEGER_TYPE:      unmarshal_int,
                 INTEGER32_TYPE:    unmarshal_int,
                 LONG_TYPE:         unmarshal_long,
                 DOUBLE_TYPE:       unmarshal_double,
                 UUID_TYPE:         unmarshal_uuid,
                 LEXICAL_UUID_TYPE: unmarshal_uuid,
                 TIME_UUID_TYPE:    unmarshal_uuid,
                 TIMESTAMP_TYPE:    unmarshal_timestamp,
                 NEW_TIMESTAMP_TYPE: unmarshal_timestamp,
                 COUNTER_TYPE:      unmarshal_initializable_int,
                 LIST_TYPE:         unmarshal_list,
                 SET_TYPE:          unmarshal_set,
                 MAP_TYPE:          unmarshal_map}


def decode_bigint(term):
    val = int(term.encode('hex'), 16)
    if (ord(term[0]) & 128) != 0:
        val = val - (1 << (len(term) * 8))
    return val


def __escape_quotes(term):
    assert isinstance(term, (str, unicode))
    return term.replace("\'", "''")



import os
from setuptools import setup


def getPackages(base):
    packages = []

    def visit(arg, directory, files):
        if '__init__.py' in files:
            packages.append(directory.replace('/', '.'))

    os.path.walk(base, visit, None)

    return packages


setup(
    name='silverberg',
    version='0.1.12',
    description='Twisted CQL Cassandra Client',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Framework :: Twisted'
    ],
    maintainer='Ken Wronkiewicz',
    maintainer_email='ken.wronkiewicz@rackspace.com',
    license='APL2',
    url='https://github.com/rackerlabs/silverberg/',
    long_description=open('README.rst').read(),
    packages=getPackages('silverberg'),
    install_requires=[
        'Twisted >= 12.0.0',
        'thrift == 0.9.0',
        'mock >= 0.9',
        'cql >= 1.4.0'
    ],
)
